/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2012
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2012
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file I_DP2PMatch.h
 *       @see I_DP2PMatch.
 *
 *  @date 20.01.2012
 *  @author Tobias Hilbrich, Mathias Korepkat, Joachim Protze, Fabian Hnsel
 */

#include "I_Module.h"
#include "GtiEnums.h"
#include "BaseIds.h"
#include "I_ChannelId.h"
#include "MustEnums.h"
#include "I_Comm.h"

#ifndef I_DP2PMATCH_H
#define I_DP2PMATCH_H

/**
 * Interface for a point to point message matching.
 *
 * Suspension and operation queuing is implemented
 * with I_OperationReordering and is also used to impose
 * orderings relating from analyzing only feasible operations
 * for deadlock detection.
 *
 * Dependencies (order as listed):
 * - ParallelIdAnalysis
 * - LocationAnalysis
 * - BaseConstants
 * - CreateMessage  ##for printing lost messages
 * - CommTrack
 * - RequestTrack
 * - DatatypeTrack
 * - FloodControl
 *
 */
class I_DP2PMatch : public gti::I_Module
{
public:

    /**
     * Notification of MPI_Init.
     *
     * @param pId parallel Id of the call site.
     * @return see gti::GTI_ANALYSIS_RETURN.
     */
    virtual gti::GTI_ANALYSIS_RETURN init (
            MustParallelId pId
    ) = 0;

    /**
     * Notification of a started send.
     *
     * @param pId parallel Id of the call site.
     * @param lId location Id of the call site.
     * @param dest to send to.
     * @param tag of the send.
     * @param comm for the send.
     * @param type for the send.
     * @param count of the send.
     * @return see gti::GTI_ANALYSIS_RETURN.
     */
    virtual gti::GTI_ANALYSIS_RETURN send (
            MustParallelId pId,
            MustLocationId lId,
            int dest,
            int tag,
            MustCommType comm,
            MustDatatypeType type,
            int count,
            int mode
    ) = 0;

    /**
     * Notification of a started non-blocking send.
     *
     * @param pId parallel Id of the call site.
     * @param lId location Id of the call site.
     * @param dest to send to.
     * @param tag of the send.
     * @param comm for the send.
     * @param type for the send.
     * @param count of the send.
     * @param request of the isend.
     * @return see gti::GTI_ANALYSIS_RETURN.
     */
    virtual gti::GTI_ANALYSIS_RETURN isend (
            MustParallelId pId,
            MustLocationId lId,
            int dest,
            int tag,
            MustCommType comm,
            MustDatatypeType type,
            int count,
            int mode,
            MustRequestType request
    ) = 0;

    /**
     * Motification of a started recv.
     *
     * @param pId parallel Id of the call site.
     * @param lId location Id of the call site.
     * @param source to receive from.
     * @param tag of the receive.
     * @param comm for the receive.
     * @param type for the recv.
     * @param count of the recv.
     * @return see gti::GTI_ANALYSIS_RETURN.
     */
    virtual gti::GTI_ANALYSIS_RETURN recv (
            MustParallelId pId,
            MustLocationId lId,
            int source,
            int tag,
            MustCommType comm,
            MustDatatypeType type,
            int count) = 0;

    /**
     * Notification of a non-blocking receive.
     *
     * @param pId parallel Id of the call site.
     * @param lId location Id of the call site.
     * @param source to receive from.
     * @param tag of the receive.
     * @param comm for the receive.
     * @param type for the recv.
     * @param count of the recv.
     * @param request for the non-blocking receive.
     * @return see gti::GTI_ANALYSIS_RETURN.
     */
    virtual gti::GTI_ANALYSIS_RETURN irecv (
            MustParallelId pId,
            MustLocationId lId,
            int source,
            int tag,
            MustCommType comm,
            MustDatatypeType type,
            int count,
            MustRequestType request) = 0;

    /**
     * Provides the actual source for a completed blocking wildcard receive.
     *
     * @param pId parallel Id of the call site.
     * @param lId location Id of the call site.
     * @param source for the receive.
     * @return see gti::GTI_ANALYSIS_RETURN.
     */
    virtual gti::GTI_ANALYSIS_RETURN recvUpdate (
            MustParallelId pId,
            MustLocationId lId,
            int source) = 0;

    /**
     * Provides the actual source for a completed non-blocking
     * wildcard receive.
     *
     * @param pId parallel Id of the call site.
     * @param lId location Id of the call site.
     * @param source that actualy was used in the wildcard receive.
     * @return see gti::GTI_ANALYSIS_RETURN.
     */
    virtual gti::GTI_ANALYSIS_RETURN irecvUpdate (
            MustParallelId pId,
            MustLocationId lId,
            int source,
            MustRequestType request) = 0;

    /**
     * Notification of a started persistent send/recv operaiton.
     *
     * @param pId parallel Id of the call site.
     * @param lId location Id of the call site.
     * @param request that was started.
     * @return see gti::GTI_ANALYSIS_RETURN.
     */
    virtual gti::GTI_ANALYSIS_RETURN startPersistent (
            MustParallelId pId,
            MustLocationId lId,
            MustRequestType request) = 0;

    /**
     * Notification of a cancel operation.
     *
     * May require to undo matches, very hard to support.
     *
     * @param pId parallel Id of the call site.
     * @param lId location Id of the call site.
     * @param request to cancel.
     * @return see gti::GTI_ANALYSIS_RETURN.
     */
    virtual gti::GTI_ANALYSIS_RETURN cancel (
            MustParallelId pId,
            MustLocationId lId,
            MustRequestType request) = 0;

    /**
     * Prints all unmatched send/recv operations if this
     * action was issued by all connected places.
     *
     * @param thisChannel channel id of process or set of completed processes
     *               triggering this action.
     * @return see gti::GTI_ANALYSIS_RETURN.
     */
    virtual gti::GTI_ANALYSIS_RETURN printLostMessages (
            gti::I_ChannelId *thisChannel) = 0;

};/*class I_DP2PMatch*/

#endif /*I_DP2PMATCH_H*/
